/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: helpfuncs.c,v 1.4 2007/08/31 08:56:13 piru Exp $
*/

#define USE_BOOPSI_STUBS
#include <aros/macros.h>
#include <datatypes/datatypes.h>
#include <proto/utility.h>
#include <proto/dos.h>
#include <proto/iffparse.h>
#include <proto/intuition.h>
#include <proto/locale.h>
#include <utility/name.h>
#include <intuition/cghooks.h>
#include <libraries/locale.h>
#include <datatypes/datatypesclass.h>
#include "datatypes_intern.h"
#ifndef __MORPHOS__
#include <clib/boopsistubs.h>
#else
#include <clib/alib_protos.h>
#endif

#undef DEBUG
#define DEBUG 0
#include <aros/debug.h>

/************************** ASCII/BINARY RECOGNITION *************************/

static
UBYTE const ASCIITable[256]=
{
   0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,0,1,0,0,0,0,0,0,0,1,0,0,0,0,
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};


static
UBYTE const BinaryTable[256]=
{
   1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,0,1,0,1,1,1,1,1,1,1,0,1,1,1,1,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};



/*
 * This works only safely because adddatatype opens datatypes.lib before
 * and therefore the init is always done before adddatatype uses the named objects.
 * Also works without adddatatype then
 */
struct DataTypesList *GetDataTypesList(struct DataTypesBase *DataTypesBase)
{
    struct NamedObject   *no;
    struct DataTypesList *dtl = NULL;

    if((no = FindNamedObject(NULL, DATATYPESLIST, NULL)))
    {
	ReleaseNamedObject(no);
	dtl = (struct DataTypesList *)no->no_Object;
    }

    if(!dtl)
    {
	if((no = AllocNamedObject(DATATYPESLIST,
				  ANO_NameSpace, TRUE,
				  ANO_UserSpace, sizeof(struct DataTypesList),
				  ANO_Flags    , NSF_NODUPS | NSF_CASE,
				  TAG_DONE)))
	{
	    D(bug("%s: created NamedObject 0x%lx\n",__func__,no));
	    if (!(dtl = (struct DataTypesList*)no->no_Object))
	    {
		FreeNamedObject(no);
		no = NULL;
	    }
	}

	if(dtl)
	{

	    D(bug("%s: dtl 0x%lx\n",__func__,dtl));

	    InitSemaphore(&dtl->dtl_Lock);

	    NewList(&dtl->dtl_SortedList);
	    NewList(&dtl->dtl_BinaryList);
	    NewList(&dtl->dtl_ASCIIList);
	    NewList(&dtl->dtl_IFFList);
	    NewList(&dtl->dtl_MiscList);

	    if(!AddNamedObject(NULL, no))
	    {
		FreeNamedObject(no);
		no = NULL;
		dtl = NULL;
	    }
	}
    }

    return(dtl);
}


struct Node *FindNameNoCase(struct Library *DataTypesBase, struct List *list,
			    STRPTR name)
{
    struct Node *node;
    struct Node *result = NULL;

    for (node = list->lh_Head; node->ln_Succ; node = node->ln_Succ)
    {
	if (!Stricmp(node->ln_Name, name))
	{
	    result = node;
	    break;
	}
    }

    return result;
}



BPTR NewOpen(struct Library *DataTypesBase, BPTR lock, STRPTR name)
{
    BPTR returnfh = NULL;
    //    struct XpkFib *xpkfib=NULL;
    BPTR dosfile = NULL;

    D(bug("datatypes.library/NewOpen: name = %s\n", name));

    if (lock)
    {
	BPTR tmplock = DupLock(lock);
	if (tmplock)
	{
	    dosfile = OpenFromLock(tmplock);
	    if (!dosfile)
	    {
		UBYTE namebuf[510];

		UnLock(tmplock);

		D(bug("datatypes.library/NewOpen: could not OpenFromLock, try NameFromLock + Open\n"));

		if (NameFromLock(lock, namebuf, sizeof(namebuf)))
		{
		    D(bug("datatypes.library/NewOpen: NameFromLock okay. Name = \"%s\"\n", namebuf));
		    dosfile = Open(namebuf, MODE_OLDFILE);
		}
	    }
	}
    }

    if(!dosfile && name)
    {
	dosfile = Open(name, MODE_OLDFILE);
    }

    if(dosfile)
    {
	D(bug("datatypes.library/NewOpen: open okay\n"));

	returnfh = dosfile;

#if 0
	if(XpkBase)
	{
	    if(xpkfib = AllocVec(sizeof(struct XpkFib),
				  MEMF_PUBLIC|MEMF_CLEAR))
	    {
		if(!xpkexaminetags(DataTypesBase, xpkfib, XPK_InFH, dosfile,
				   TAG_DONE))
		{
		    switch (xpkfib->Type)
		    {
		    case XPKTYPE_UNPACKED:
			returnfh=dosfile;
			break;

		    case XPKTYPE_PACKED:
			Message(DataTypesBase,"file is XPK packed","okay");

			if (xpkfib->Flags&XPKFLAGS_PASSWORD)
			    Message(DataTypesBase,"file needs password","okay");

			if (xpkfib->Flags&XPKFLAGS_NOSEEK)
			    Message(DataTypesBase,"file does not support seeking",
				    "okay");
			if (xpkfib->Flags&XPKFLAGS_NONSTD)
			    Message(DataTypesBase,"file is non standard","okay");

			SetIoErr(ERROR_NOT_IMPLEMENTED);
			break;

		    case XPKTYPE_ARCHIVE:
			SetIoErr(ERROR_NOT_IMPLEMENTED);
			break;
		    }
		}
		FreeVec(xpkfib);
	    }
	}
	if (returnfh != dosfile)
	{
	    Close(dosfile);
	}
#endif

    }

    return returnfh;
}




struct CompoundDatatype *ExamineLock(struct Library *DataTypesBase, struct CompoundDatatype *CurrentDT, ULONG GroupID, BPTR lock, struct FileInfoBlock *fib)
{
    struct CompoundDatatype *cdt = NULL;

    D(bug("datatypes.library/ExamineLock\n"));

    if(Examine(lock, fib))
    {
        D(bug("datatypes.library/ExamineLock: Examine okay\n"));

	ObtainSemaphoreShared(&getDTLIST->dtl_Lock);

	if (fib->fib_DirEntryType > 0)
	{
	    D(bug("datatypes.library/ExamineLock: is a directory\n"));
	    cdt = (struct CompoundDatatype *)FindNameNoCase(DataTypesBase,
							    &getDTLIST->dtl_MiscList,
							    "directory");
	}
	else
	{
	    //if (fib->fib_DirEntryType < 0)
	    {
		BPTR file;

		D(bug("datatypes.library/ExamineLock: is a file\n"));

		if((file = NewOpen(DataTypesBase, lock, NULL)))
		{
		    UBYTE *CheckArray;
		    UWORD CheckSize = (getDTLIST->dtl_LongestMask > 64) ?
		                       getDTLIST->dtl_LongestMask : 64;

		    D(bug("datatypes.library/ExamineLock: NewOpen okay\n"));

		    if((CheckArray = AllocVec((ULONG)(CheckSize)+1,
		                              MEMF_ANY)))
		    {
			D(bug("datatypes.library/ExamineLock: Allocated CheckArray\n"));

			if((CheckSize = Read(file, CheckArray,
			                     (ULONG)CheckSize)) > 0)
			{
			    struct DTHookContext dthc;
			    struct IFFHandle *iff;

			    D(bug("datatypes.library/ExamineLock: Read in CheckArray size = %d\n", CheckSize));

			    CheckArray[CheckSize] = '\0';

			    Seek(file, 0, OFFSET_BEGINNING);

			    dthc.dthc_SysBase = (struct Library *)SysBase;
			    dthc.dthc_DOSBase = DOSBase;
			    dthc.dthc_IFFParseBase = IFFParseBase;
			    dthc.dthc_UtilityBase = UtilityBase;
			    dthc.dthc_Lock = lock;
			    dthc.dthc_FIB = fib;
			    dthc.dthc_FileHandle = file;
			    dthc.dthc_Buffer = CheckArray;
			    dthc.dthc_BufferLength = CheckSize;

			    if(!(iff=dthc.dthc_IFF = AllocIFF()))
				SetIoErr(ERROR_NO_FREE_STORE);
			    else
			    {
				D(bug("datatypes.library/ExamineLock: AllocIFF okay: iff = %x\n", iff));

				iff->iff_Stream = (IPTR)file;
				InitIFFasDOS(iff);

				if (!OpenIFF(iff, IFFF_READ))
				{
				    D(bug("datatypes.library/ExamineLock: OpenIFF okay. Now calling ExamineData\n"));

				    cdt = ExamineData(DataTypesBase,
				                      &dthc,
				                      CurrentDT,
				                      GroupID,
				                      CheckArray,
				                      CheckSize,
				                      fib->fib_FileName,
				                      fib->fib_Size);

				    D(bug("datatypes.library/ExamineLock: ExamineData() returned %x\n", cdt));

				    CloseIFF(iff);

				} /* OpenIFF okay */

				FreeIFF(iff); /* AROS BUG FIX: was dthc.dthc_IFF) */

			    } /* AllocIFF okay */

			} /* if (CheckSize = Read(... */

			FreeVec(CheckArray);

		    } /* if (CheckArray = AllocVec(... */

		    Close(file);

		} /* if file opened */

	    } /* it is a file */

	} /* it is not a directory */

	ReleaseSemaphore(&getDTLIST->dtl_Lock);

    } /* if(Examine(lock, fib)) */

    return cdt;
}


/* Simple IFF hook to read from mystream handle - Piru */

struct mystream
{
    struct Library *DataTypesBase;
    CONST_APTR      Buffer;
    ULONG           BufSize;
    LONG            Pos;
};

static
LONG memhook_code(struct Hook *streamhook,
                  struct IFFHandle *iff,
                  struct IFFStreamCmd *cmd)
{
    struct mystream *stream = (APTR) iff->iff_Stream;
    struct Library *DataTypesBase = stream->DataTypesBase;
    LONG ret = 0;

    switch (cmd->sc_Command)
    {
	case IFFCMD_INIT:
	    D(bug("datatypes.library/memhook_code: IFFCMD_INIT\n"));
	    break;

	case IFFCMD_CLEANUP:
	    D(bug("datatypes.library/memhook_code: IFFCMD_CLEANUP\n"));
	    stream->Pos = 0;
	    break;

	case IFFCMD_READ:
	    {
		LONG actual;

		D(bug("datatypes.library/memhook_code: IFFCMD_READ %ld bytes to 0x%lx\n", cmd->sc_NBytes, cmd->sc_Buf));

		if (stream->Pos < 0 || stream->Pos > stream->BufSize)
		{
		    D(bug("datatypes.library/memhook_code: IFFCMD_READ file position bogus, fail\n"));
		    ret = -1;
		    break;
		}

		actual = stream->BufSize - stream->Pos;
		if (cmd->sc_NBytes > actual)
		{
		    D(bug("datatypes.library/memhook_code: IFFCMD_READ would overflow, fail\n"));
		    ret = -1;
		    break;
		}

		CopyMem(((UBYTE *) stream->Buffer) + stream->Pos,
		        cmd->sc_Buf,
		        cmd->sc_NBytes);

		stream->Pos += cmd->sc_NBytes;

		D(bug("datatypes.library/memhook_code: IFFCMD_READ read %ld bytes\n", cmd->sc_NBytes));
	    }
	    break;

	case IFFCMD_WRITE:
	    D(bug("datatypes.library/memhook_code: IFFCMD_WRITE: not supported, fail\n"));
	    ret = -1;
	    break;

	case IFFCMD_SEEK:
	    {
		LONG newpos = stream->Pos + cmd->sc_NBytes;

		D(bug("datatypes.library/memhook_code: IFFCMD_SEEK by %ld\n", cmd->sc_NBytes));

		if (newpos < 0 || newpos > stream->BufSize)
		{
		    D(bug("datatypes.library/memhook_code: IFFCMD_SEEK new file position is bogus, fail\n"));
		    ret = -1;
		    break;
		}

		stream->Pos = newpos;

		D(bug("datatypes.library/memhook_code: IFFCMD_SEEK okay: newpos = %ld\n", newpos));
	    }
	    break;

	default:
	    D(bug("datatypes.library/memhook_code: IFFCMD_??? %ld\n", cmd->sc_Command));
	    ret = -1;
	    break;
    }

    return ret;
}

static const struct Hook memhook =
{
    {0, 0},
    (HOOKFUNC) &HookEntry,
    (HOOKFUNC) memhook_code,
    NULL
};


struct CompoundDatatype *ExamineMemory(struct Library *DataTypesBase, struct CompoundDatatype *CurrentDT, ULONG GroupID, CONST_APTR SourceAddress, ULONG SourceSize)
{
    struct CompoundDatatype *cdt = NULL;

    D(bug("datatypes.library/ExamineMemory\n"));

    if (SourceAddress == NULL || SourceSize == 0)
	SetIoErr(ERROR_OBJECT_NOT_FOUND);
    else
    {
	struct DTHookContext dthc;
	struct IFFHandle *iff = NULL;
	UWORD CheckSize;
	struct mystream handle;

	ObtainSemaphoreShared(&getDTLIST->dtl_Lock);

	CheckSize = (getDTLIST->dtl_LongestMask > 64) ?
	             getDTLIST->dtl_LongestMask : 64;

	if (SourceSize < CheckSize)
	{
	    CheckSize = SourceSize;
	}

	dthc.dthc_SysBase = (struct Library *)SysBase;
	dthc.dthc_DOSBase = DOSBase;
	dthc.dthc_IFFParseBase = IFFParseBase;
	dthc.dthc_UtilityBase = UtilityBase;
	dthc.dthc_Lock = NULL;
	dthc.dthc_FIB = NULL;
	dthc.dthc_FileHandle = NULL;
	dthc.dthc_Buffer = (APTR) SourceAddress;
	dthc.dthc_BufferLength = SourceSize;

	/* If the buffer appears to hold an IFF object, feed it as IFFHandle,
	 * else as RAW data. I wish there was a bit more elegant way of
	 * handling this... - Piru
	 */
	if (CheckSize >= 8)
	{
	    ULONG IFF_ID   = AROS_BE2LONG(((ULONG*)SourceAddress)[0]);
	    ULONG IFF_Size = AROS_BE2LONG(((ULONG*)SourceAddress)[1]);

	    if (SourceSize * 3 / 4 < IFF_Size && SourceSize * 4 / 3 > IFF_Size &&
	        (IFF_ID == ID_FORM || IFF_ID == ID_CAT || IFF_ID == ID_LIST))
	    {
		D(bug("datatypes.library/ExamineMemory: Appears to be IFF data\n"));

		if((iff = AllocIFF()))
		{
		    D(bug("datatypes.library/ExamineMemory: AllocIFF okay: iff = %x\n", iff));

		    handle.DataTypesBase = DataTypesBase;
		    handle.Buffer        = SourceAddress;
		    handle.BufSize       = CheckSize;
		    handle.Pos           = 0;

		    iff->iff_Stream = (IPTR) &handle;
		    InitIFF(iff, IFFF_FSEEK | IFFF_RSEEK, &memhook);

		    if (OpenIFF(iff, IFFF_READ) != 0)
		    {
			D(bug("datatypes.library/ExamineMemory: OpenIFF failed\n"));

			FreeIFF(iff);
			iff = NULL;
		    }
		}
	    }
	}

	/* 'iff' is either NULL or the IFFhandle */
	dthc.dthc_IFF = iff;

	cdt = ExamineData(DataTypesBase,
	                  &dthc,
	                  CurrentDT,
	                  GroupID,
	                  (APTR) SourceAddress,
	                  CheckSize,
	                  "",
	                  SourceSize);

	D(bug("datatypes.library/ExamineMemory: ExamineData() returned %x\n", cdt));

	ReleaseSemaphore(&getDTLIST->dtl_Lock);

	if (iff)
	{
	    CloseIFF(iff);
	    FreeIFF(iff);
	}
    }

    return cdt;
}


struct CompoundDatatype *ExamineData(struct Library *DataTypesBase,
				     struct DTHookContext *dthc,
				     struct CompoundDatatype *CurrentDT,
				     ULONG GroupID,
				     UBYTE *CheckArray, UWORD CheckSize,
				     UBYTE *Filename, ULONG Size)
{
    struct CompoundDatatype *cdt = NULL;

    UWORD        type;
    struct List *list  = NULL;
    BOOL         found = FALSE;

    ULONG IFF_ID = 0;
    ULONG IFF_Size = 0;

    D(bug("%s: DTHookContext 0x%lx CheckArray 0x%lx CheckSize %ld FileName <%s> Size %ld\n",
	__func__,
	dthc,
	CheckArray,
	CheckSize,
	Filename,
	Size));

    if (CheckSize >= 8)
    {
	IFF_ID   = AROS_BE2LONG(*((ULONG*)CheckArray));
	IFF_Size = AROS_BE2LONG(*((ULONG*)(CheckArray+4)));
    }

    if(((!dthc->dthc_FileHandle) && (dthc->dthc_IFF)) ||
       (((Size*3/4 < IFF_Size) && (Size*4/3 > IFF_Size)) &&
	(IFF_ID==ID_FORM || IFF_ID==ID_CAT || IFF_ID==ID_LIST)) )
    {
	D(bug("%s: DTF_IFF\n", __func__));
	type = DTF_IFF;
	list = &getDTLIST->dtl_IFFList;
    }
    else
    {
	UBYTE *ptr;
	UWORD count;
	UWORD ascii;

	dthc->dthc_IFF = NULL;

	for (ptr=CheckArray,count=CheckSize,ascii=0 ; count-- ; ptr++)
	{
	    if (ASCIITable[*ptr])
	    {
		ascii++;
	    }
	    else
	    {
		if (BinaryTable[*ptr])
		{
		    ascii=0;
		    break;
		}
	    }
	}

	if (ascii > CheckSize*3/4)
	{
	    D(bug("%s: DTF_ASCII\n", __func__));
	    type = DTF_ASCII;
	    list = &getDTLIST->dtl_ASCIIList;
	}
	else
	{
	    D(bug("%s: DTF_BINARY\n", __func__));
	    type = DTF_BINARY;
	    list = &getDTLIST->dtl_BinaryList;
	}
    }

    if (list)
    {
	struct CompoundDatatype *cur;

	for(cur = (CurrentDT ? (struct CompoundDatatype *)CurrentDT->DT.dtn_Node1.ln_Succ : (struct CompoundDatatype *)list->lh_Head);
	    cur && cur->DT.dtn_Node1.ln_Succ;
	    cur = (struct CompoundDatatype *)cur->DT.dtn_Node1.ln_Succ)
	{
	    D(bug("%s: current CompoundDataType 0x%lx\n",__func__,cur));
	    D(bug("groupid  = %c%c%c%c\n",
			 cur->DTH.dth_GroupID >> 24,
			 cur->DTH.dth_GroupID >> 16,
			 cur->DTH.dth_GroupID >> 8,
			 cur->DTH.dth_GroupID));
	    D(bug("id       = %c%c%c%c\n",
			 cur->DTH.dth_ID >> 24,
			 cur->DTH.dth_ID >> 16,
			 cur->DTH.dth_ID >> 8,
			 cur->DTH.dth_ID));
	    D(bug("flags    = %x\n",	 cur->DTH.dth_Flags));
	    D(bug("pri      = %d\n",	 cur->DTH.dth_Priority));
	    D(bug("name     = %s\n",	 cur->DTH.dth_Name));
	    D(bug("basename = %s\n",	 cur->DTH.dth_BaseName));
	    D(bug("pattern  = %s\n",	 cur->DTH.dth_Pattern));
	    D(bug("masklen  = %d\n",	 cur->DTH.dth_MaskLen));

	    if (GroupID && cur->DTH.dth_GroupID != GroupID) continue;

	    if( CheckSize >= cur->DTH.dth_MaskLen)
	    {
		WORD *msk = cur->DTH.dth_Mask;
		UBYTE *cmp = CheckArray;
		UWORD count;

		found=TRUE;

		D(bug("%s: MaskLen %ld\n",__func__,cur->DTH.dth_MaskLen));

		for (count = cur->DTH.dth_MaskLen; count--; msk++, cmp++)
		{
		    D(bug("%s: *msk 0x%lx\n",__func__,*msk));
		    if (*msk>0)
		    {
			if(cur->DTH.dth_Flags & DTF_CASE)
			{
			    if (*msk != *cmp)
			    {
				found=FALSE;
				break;
			    }
			}
			else
			{
			    if(*msk != *cmp &&
			       *msk != ToUpper((ULONG)*cmp) &&
			       *msk != ToLower((ULONG)*cmp))
			    {
				found=FALSE;
				break;
			    }
			}
		    }
		}

		if(found)
		{
		    if((!(cur->FlagLong & CFLGF_PATTERN_UNUSED)) &&
		       cur->DTH.dth_Pattern)
		    {
			if(cur->FlagLong & CFLGF_IS_WILD)
			{
			    if(cur->ParsePatMem)
			    {
				if(!MatchPatternNoCase(cur->ParsePatMem,
						       Filename))
				{
				    found = FALSE;
				}
			    }
			}
			else
			{
			    if(Stricmp(cur->DTH.dth_Pattern, Filename))
			    {
				found = FALSE;
			    }
			}
		    }

		    if(found)
		    {
			D(bug("%s: check function 0x%lx\n",__func__,cur->Function));
			if(cur->Function)
			{
#ifdef __MORPHOS__
			    REG_A0	=(ULONG) dthc;
			    found	=	(*MyEmulHandle->EmulCallDirect68k)(cur->Function);
#else
			    found = (cur->Function)(dthc);
#endif

			    D(bug("%s: found 0x%lx\n",__func__,found));

			    if (dthc->dthc_IFF)
			    {
				D(bug("%s: reset iff pos\n",__func__));
				CloseIFF(dthc->dthc_IFF);
				OpenIFF(dthc->dthc_IFF, IFFF_READ);
			    }
			    else
			    {
				D(bug("%s: reset file pos\n",__func__));
				Seek(dthc->dthc_FileHandle, 0,
				     OFFSET_BEGINNING);
			    }
			}
		    }
		}
	    }
	    else
	    {
		D(bug("%s: file too short to check\n", __func__));
	    }

	    if(found)
	    {
		cdt = cur;
		break;
	    }
	}

	if(!found && !CurrentDT && !GroupID)
	{
	    D(bug("%s: not found, take default type %ld\n",__func__,type));
	    switch(type)
	    {
	    case DTF_BINARY:
		cdt = (struct CompoundDatatype *)FindNameNoCase(DataTypesBase, list,
								"binary");
		break;

	    case DTF_ASCII:
		cdt = (struct CompoundDatatype *)FindNameNoCase(DataTypesBase, list,
							       "ascii");
		break;

	    case DTF_IFF:
		cdt = (struct CompoundDatatype *)FindNameNoCase(DataTypesBase, list,
								"iff");
		break;
	    }
	}
    }

    return cdt;
}


#undef getDTLIST


#ifndef __MORPHOS__
/* Putchar procedure needed by RawDoFmt() */

AROS_UFH2(void, dt_putchr,
	  AROS_UFHA(UBYTE,    chr, D0),
	  AROS_UFHA(STRPTR *, p,   A3))
{
    AROS_USERFUNC_INIT
    *(*p)++ = chr;
    AROS_USERFUNC_EXIT
}

void dt__sprintf(struct Library *DataTypesBase, UBYTE *buffer,
		 UBYTE *format, ...)
{
    RawDoFmt(format, &format+1, (VOID_FUNC)dt_putchr, buffer);
}
#else

void dt_sprintf(struct Library *DataTypesBase, UBYTE *buffer, UBYTE *format, ...)
{
    va_list args;

    va_start(args,format);

    VNewRawDoFmt((CONST_STRPTR) format, (APTR (*)(APTR, UBYTE)) 0, buffer, args);

    va_end(args);
}

#endif

#warning these work only with stack growing downwards and should therefore be fixed to use macros in utility/tagitem.h

ULONG VARARGS	setattrs(struct Library *DataTypesBase, Object *object, ...)
{
    va_list args;
    ULONG ret;
    va_start(args, object);
    ret = SetAttrsA(object, (struct TagItem *)GET_VALIST_PTR(args));
    va_end(args);
    return ret;
}


ULONG VARARGS	Do_OM_NOTIFY(struct Library *DataTypesBase, Object *object,
		   struct GadgetInfo *ginfo, ULONG flags, ...)
{
    va_list args;
    ULONG ret;
    struct opUpdate opu;

    va_start(args, flags);

    opu.MethodID = OM_NOTIFY;
    opu.opu_AttrList = (struct TagItem *)GET_VALIST_PTR(args);
    opu.opu_GInfo = ginfo;
    opu.opu_Flags = flags;

    ret = DoMethodA(object, (Msg)&opu);

    va_end(args);

    return ret;
}


ULONG VARARGS	DoGad_OM_NOTIFY(struct Library *DataTypesBase, Object *object,
		      struct Window *win, struct Requester *req,
		      ULONG flags, ...)
{
    va_list args;
    ULONG ret;
    va_start(args, flags);
    ret = dogadgetmethod(DataTypesBase, (struct Gadget*)object, win, req, OM_NOTIFY,
			  GET_VALIST_PTR(args), NULL, flags);
    va_end(args);
    return ret;
}


ULONG VARARGS	dogadgetmethod(struct Library *DataTypesBase, struct Gadget *gad,
		     struct Window *win, struct Requester *req, ...)
{
    va_list args;
    ULONG ret;
    va_start(args, req);
    ret = DoGadgetMethodA(gad, win, req, (Msg)GET_VALIST_PTR(args));
    va_end(args);
    return ret;
}


struct Catalog * VARARGS	opencatalog(struct Library *DataTypesBase, struct Locale *locale,
			    STRPTR name, ...)
{
    struct Catalog *ret = NULL;

    if (LocaleBase)
    {
	va_list args;
	va_start(args, name);
	ret = OpenCatalogA(locale, name, (struct TagItem *)GET_VALIST_PTR(args));
	va_end(args);
    }

    return ret;
}
